import sys
from enum import Enum
from struct import unpack_from
from math import ceil

MAXCOLUMNS = 156
MAXLINES = 32

class GenStyle(Enum):
	GPRINT = 1
	DATA   = 2
	C      = 3

GEN_STYLE = GenStyle.C

BAS_FIRSTLINENR = 10000
BAS_LINEINCR = 10
BAS_MAXBYTES_PER_GPRINT = 28
BAS_MAXBYTES_PER_DATA = 10
C_MAXBYTES_PER_LINE = 20


BM = 19778 # value of "BM", marker for Windows bitmap file type
verbose = True;


# reads bitmap from file
def readBitmap(fileName):
	# read raw content of input file into memory
	f = open(fileName,'rb')
	rawContent=bytearray(f.read())
	f.close()

	# parse bitmap header
	type=unpack_from('H',rawContent,0)[0] # unsigned short, must be "BM"
	pxWidth=unpack_from('<l',rawContent,18)[0] # long, little endian
	pxHeight=unpack_from('<l',rawContent,22)[0] # long, little endian
	numLines = abs(pxHeight); # negative pxHeight indicates top-down bitmap
	depth=unpack_from('H',rawContent,28)[0] # unsigned short, should be 1
	compression=unpack_from('H',rawContent,30)[0] # unsigned short, should be 0
	offsetData=unpack_from('<L',rawContent,10)[0] # offset where bitmap data begins

	if pxHeight > 0:
		bottomUp = True
	else:
		bottomUp = False

	if verbose:
		# report header data
		print("file type:", type)
		print("pxWidth:", pxWidth)
		print("pxHeight:", pxHeight)
		print("color depth:", depth)
		print("compression:", compression)
		if bottomUp:
			print("bottom-up")
		else:
			print("top-down")
		print()
	 
	# constraint check
	if type != BM: raise Exception("unsupported file type - only Windows bitmaps accepted")
	if depth != 1: raise Exception("can only process bitmaps with color depth 1")
	if compression: raise Exception("can only process uncompressed bitmaps")

	# compute lineSizeBytes - all lines are 4 byte aligned
	numUsedBytes = ceil(pxWidth / 8)
	numFillBytes = (4 - numUsedBytes % 4 ) % 4 # up to 3 fill bytes for 4 byte alignment
	lineSizeBytes = numUsedBytes + numFillBytes

	# check for bottom-up vs. top-down bitmap
	if bottomUp:
	 lineRange = range(numLines-1,-1,-1)
	else:
	 lineRange = range(0,numLines)

	# extract all bits line-wise and store them in a bit-matrix
	# i.e. an associative map with (y,x)-tupels as index
	# allways top-down - i.e. Matrix[0][.] represents the top line
	Matrix = {} 
	curLineIdxForMatrix = 0
	for curLine in lineRange:
	 lineOffset = offsetData + curLine * lineSizeBytes
	 for curPxIdx in range(0, pxWidth):
	  byteOffset = int(curPxIdx/8)
	  curByte = rawContent[lineOffset + byteOffset]
	  curBitIdxInByte = curPxIdx % 8
	  curBit = (1 << (7 - curBitIdxInByte)) & curByte
	  if curBit:
	    Matrix[curLineIdxForMatrix,curPxIdx] = 1
	  else:		
	    Matrix[curLineIdxForMatrix,curPxIdx] = 0
	 curLineIdxForMatrix += 1
	return numLines, pxWidth, Matrix

	
# print bitmap to stdout - for debug purposes
def showBitmap(bitmap, lines, columns):
	for curLine in range(0,lines):
	 for curCol in range(0,columns):
	  if bitmap[curLine,curCol]:
	    print(" ",end='')
	  else:
	    print("0",end='')
	 print()
	print()

# print bitmap content to stdout - for debug purposes
def printBits(bitmap, lines, columns):
	for curLine in range(0,lines):
	 for curCol in range(0,columns):
	  print(bitmap[curLine,curCol], end='')
	 print()
	print()

	
# code generator
class CodeGenerator:
	halfByteToHexDigit = ["0","1","2","3","4","5","6","7","8","9","A","B","C","D","E","F"]
	
	def __init__(self, bitmap, lines, columns, startX, startY, genStyle, outfile):
		self.bitmap = bitmap
		self.numLines = lines
		self.numCols = columns
		self.startX = startX
		self.startY = startY
		self.curBasLineNr = BAS_FIRSTLINENR
		self.out = outfile
		self.curLine = 0
		self.curCol  = 0
		self.curBytesInStatement = 0
		self.genStyle = genStyle
	
	def writeLineNr(self):
		if self.genStyle == GenStyle.GPRINT or self.genStyle == GenStyle.DATA:
			self.out.write(str(self.curBasLineNr) + " ")
			self.curBasLineNr += BAS_LINEINCR
	
	def curLineNr(self):
		return str(self.curBasLineNr - BAS_LINEINCR) # lineNr has been allready incremented
	
	def writeEOL(self):
		self.out.write('\r\n')
	
	def writeEOF(self):
		if self.genStyle != GenStyle.C:
			self.out.write(chr(26))
		
	def writeHeader(self):
		self.writeLineNr()
		if self.genStyle == GenStyle.GPRINT:
			self.out.write("X=" + str(self.startX) + ":Y=" + str(self.startY))
		elif self.genStyle == GenStyle.DATA:
			self.out.write("REM DATA")
		self.writeEOL()
	
	def writeFooter(self):
		if self.genStyle == GenStyle.GPRINT:
			self.writeLineNr()
			self.out.write("RETURN")
			self.writeEOL()
	
	def checkNeedOfLineBreak(self):
		if self.genStyle == GenStyle.GPRINT:
			if self.curBytesInStatement >= BAS_MAXBYTES_PER_GPRINT:
				self.out.write("\"")
				self.out.write(";")
				self.writeEOL()
				self.writeLineNr()
				# self.out.write("GCURSOR X+" + str(self.curCol) + ",Y+" + str(self.curLine) + ":")
				self.out.write("GPRINT\"")
				self.curBytesInStatement = 0
		elif self.genStyle == GenStyle.DATA:
			if self.curBytesInStatement >= BAS_MAXBYTES_PER_DATA:
				self.writeEOL()
				self.writeLineNr()
				self.out.write("DATA ")
				self.curBytesInStatement = 0
		elif self.genStyle == GenStyle.C:
			if self.curBytesInStatement >= C_MAXBYTES_PER_LINE:
				self.out.write(",")
				self.writeEOL()
				self.curBytesInStatement = 0
		
				
		
	def copyBitmapBitToHalfByte(self, halfbyte, bitIdx, x, y):
		try:
			halfbyte += (1 << bitIdx) *	self.bitmap[y,x]
		except: # index out of range => no bit set
			NONE
		finally:
			return halfbyte
		
	
	def nextByte(self):
		self.checkNeedOfLineBreak()
		if self.genStyle == GenStyle.DATA or self.genStyle == GenStyle.C:
			if self.curBytesInStatement > 0:
				self.out.write(",")
			if self.genStyle == GenStyle.DATA:
				self.out.write("&")
			elif self.genStyle == GenStyle.C:
				self.out.write("0x")
				
		x = self.curCol
		y = self.curLine
		halfbyte = self.copyBitmapBitToHalfByte(0,        0, x, y+4)
		halfbyte = self.copyBitmapBitToHalfByte(halfbyte, 1, x, y+5)
		halfbyte = self.copyBitmapBitToHalfByte(halfbyte, 2, x, y+6)
		halfbyte = self.copyBitmapBitToHalfByte(halfbyte, 3, x, y+7)
		self.out.write(self.halfByteToHexDigit[halfbyte])

		halfbyte = self.copyBitmapBitToHalfByte(0,        0, x, y  )
		halfbyte = self.copyBitmapBitToHalfByte(halfbyte, 1, x, y+1)
		halfbyte = self.copyBitmapBitToHalfByte(halfbyte, 2, x, y+2)
		halfbyte = self.copyBitmapBitToHalfByte(halfbyte, 3, x, y+3)
		self.out.write(self.halfByteToHexDigit[halfbyte])
		
		self.curCol += 1
		self.curBytesInStatement += 1
			
	def nextBitmapLine(self):
		self.writeLineNr()
		self.curBytesInStatement = 0

		if self.genStyle == GenStyle.GPRINT:
			self.out.write("GCURSOR X,Y+" + str(self.curLine) + ":")
			self.out.write("GPRINT\"")
		elif self.genStyle == GenStyle.DATA:
			self.out.write("DATA ")
		while self.curCol < self.numCols:
			self.nextByte()
		if self.genStyle == GenStyle.GPRINT:
			self.out.write("\"")
		elif self.genStyle == GenStyle.C:
			if self.curLine + 8 < self.numLines:
				self.out.write(",")
		self.writeEOL()
		self.curCol   = 0
		self.curLine += 8
		return self.curLine < self.numLines 

	
# generate code for the SHARP PC-1600
def generatePC1600Code(bitmap, lines, columns, genStyle, outfile):
	# check constraints
	if lines > MAXLINES:
	  raise Exception("number of lines must not exceed ", MAXLINES)
	if columns > MAXCOLUMNS:
	  raise Exception("number of columns must not exceed ", MAXCOLUMNS)
	  
	gen = CodeGenerator(bitmap, lines, columns, 0, 0, genStyle, outfile)
	# gen.writeHeader()
	linesLeft = True
	while linesLeft:
	  linesLeft = gen.nextBitmapLine()
	gen.writeFooter()
	gen.writeEOF()

	
 
# ------------ main script ---------------

# get command-line parameters
argv = sys.argv
if len(argv) > 3 or len(argv) < 2:
  raise Exception("usage: bmp2sharp <inFile> [<outfile>]")
inFileName = argv[1]
if len(argv) == 3:
  outFile = open(argv[2],"w",newline='')
  verbose = False;
else:
  outFile = sys.stdout
  verbose = True;
 
# read .bmp file and convert it into a 2-dim bit-matrix
numLines, numColumns, bitmap = readBitmap(inFileName)
 
# process the bit-matrix
if verbose:
	showBitmap(bitmap, numLines, numColumns)
# printBits(bitmap, numLines, numColumns) # activate for debugging
generatePC1600Code(bitmap, numLines, numColumns, GEN_STYLE, outFile)

# cleanup
outFile.close()
